# -*- coding: utf-8 -*-

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '1.2.1'
__versiondate__ = '2020-05-21'
__license__ = f'Covasim {__version__} ({__versiondate__}) — © 2020 by IDM'
